/*******************************************************************************
 * Copyright (c) 2014, Juniper Networks, Inc.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Vishal Gupta (Juniper Networks) - initial implementation
 *******************************************************************************/
package org.eclipse.custom.callhierarchy.sourceview.editor;

import java.util.Iterator;

import org.eclipse.cdt.core.CCorePlugin;
import org.eclipse.cdt.core.IPositionConverter;
import org.eclipse.cdt.core.model.CModelException;
import org.eclipse.cdt.core.model.ISourceRange;
import org.eclipse.cdt.core.model.ISourceReference;
import org.eclipse.cdt.core.model.ITranslationUnit;
import org.eclipse.cdt.internal.core.model.ext.ICElementHandle;
import org.eclipse.cdt.internal.ui.callhierarchy.CHNode;
import org.eclipse.cdt.internal.ui.callhierarchy.CHReferenceInfo;
import org.eclipse.cdt.internal.ui.callhierarchy.CHViewPart;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.Region;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IEditorRegistry;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.XMLMemento;
import org.eclipse.ui.internal.EditorManager;
import org.eclipse.ui.internal.EditorReference;
import org.eclipse.ui.internal.IWorkbenchConstants;
import org.eclipse.ui.internal.WorkbenchPage;
import org.eclipse.ui.internal.WorkbenchPlugin;
import org.eclipse.ui.internal.registry.EditorDescriptor;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.ui.texteditor.ITextEditor;

/**
 * @author vishalgupta
 *
 */
@SuppressWarnings({ "restriction" })
public class SourceView extends ViewPart implements ISelectionListener{
	
	private CHNode fNavigationNode;
	private int fNavigationDetail;
	private Composite top;
	private IResource res;
	private IResource resold;
	private IEditorPart editorpart;
	private Composite parentcomposite;
	private Integer offset = 0;
	private Integer length = 0;
	private IEditorReference ref;
	private String editorId = "org.eclipse.custom.callhierarchy.sourcevieweditor";

	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createPartControl(Composite parent) {
		parentcomposite = parent;
		getSite().getWorkbenchWindow().getSelectionService().addSelectionListener("org.eclipse.cdt.ui.callHierarchy", this);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	@Override
	public void setFocus() {		

	}


	/**
	 * Create the Custom C Editor and embeds it in the SourceView
	 */
	public void createComposite() {
		// create the composite to hold the editor
		top = new Composite(parentcomposite, SWT.BORDER);
		top.setLayout(new FillLayout());
		// create the memento for the editor
		XMLMemento editorState = XMLMemento
				.createWriteRoot(IWorkbenchConstants.TAG_EDITOR_STATE);
		// Fetch the editor manager
		EditorManager manager = ((WorkbenchPage) getSite().getPage()).getEditorManager();
		// Fetch the Editor registry
		IEditorRegistry reg = WorkbenchPlugin.getDefault().getEditorRegistry();
		// Fetch the Editor descriptor
		EditorDescriptor desc = (EditorDescriptor) reg.findEditor(editorId);
		// Create the Editor reference for the Custom C Editor
		ref = new EditorReference(manager, new FileEditorInput((IFile) res),
				desc, editorState);
		// fetch the Custom C Editor
		editorpart = ref.getEditor(true);
		// embed the editor in the SourceView
		editorpart.createPartControl(top);
		// layout the editor
		parentcomposite.layout(true);
	}


	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if(selection != null){
			onShowSelectedReference(((CHViewPart)part).getTreeViewer().getSelection());			
	        showReference();
		}
	}
	
	/**
	 * Fetches the selected node
	 * @param selection on the CHView
	 */
	protected void onShowSelectedReference(ISelection selection) {
		if (selection instanceof IStructuredSelection) {
			IStructuredSelection ss= (IStructuredSelection) selection;
			for (Iterator<?> iter = ss.iterator(); iter.hasNext(); ) {
				Object cand= iter.next();
				if (cand instanceof CHNode) {
					CHNode node = (CHNode) cand;
					if (node != null && node == fNavigationNode && node.getReferenceCount() > 0) {
						fNavigationDetail= (fNavigationDetail + 1) % node.getReferenceCount();
					}
					else {
						fNavigationDetail= 0;
					}
					fNavigationNode =  (CHNode) cand;
				}
			}
		}
	}	
	
	
	/**
	 * Fetches the file from the selection on the
	 * CHView and finds the function location 
	 */
	private void showReference() {
		// if the selection on the CHView is not empty
		if (fNavigationNode != null) {
			ITranslationUnit tu = fNavigationNode.getFileOfReferences();
			if (tu != null) {
				// fetch the resource
				res = tu.getResource();
				IEditorRegistry registry = PlatformUI.getWorkbench()
						.getEditorRegistry();
				String filename = res.getName();
				filename = filename.replace(".", "#");
				String filetype = "*." + filename.split("#")[1];
				// fetch the default editor for the current selected resource
				IEditorDescriptor desc = registry.getDefaultEditor(filetype);

				IRegion region = null;
				long timestamp = 0;				
				if (fNavigationNode.getReferenceCount() > 0) {
					timestamp= fNavigationNode.getTimestamp();
					if (fNavigationDetail < 0) {
						fNavigationDetail = 0;
					} else if (fNavigationDetail >= fNavigationNode
							.getReferenceCount() - 1) {
						fNavigationDetail = fNavigationNode.getReferenceCount() - 1;
					}
					CHReferenceInfo ref = fNavigationNode
					.getReference(fNavigationDetail);
					region = new Region(ref.getOffset(), ref.getLength());
				} else {
					fNavigationDetail = 0;
					if (fNavigationNode.getRepresentedDeclaration() instanceof ISourceReference) {
						ISourceReference sr= (ISourceReference) fNavigationNode.getRepresentedDeclaration();
						tu= sr.getTranslationUnit();
						res = tu.getResource();
						ISourceRange range = null;
						try {
							range = sr.getSourceRange();
						} catch (CModelException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						if (tu.isWorkingCopy() || fNavigationNode.getRepresentedDeclaration() instanceof ICElementHandle) {
							timestamp= -1; 
						}
						region = new Region(range.getIdStartPos(), range.getIdLength());
					}
				}
				IPositionConverter pc = CCorePlugin.getPositionTrackerManager()
						.findPositionConverter(res.getFullPath(), timestamp);
				if (pc != null) {
					region = pc.historicToActual(region);
				}
				// check if the previous selection is same as the new one
				if (res != resold || fNavigationNode.getReferenceCount() > 1) {
					inputChanged(region);
					// remove CustomEEditor from the default editor list
					registry.setDefaultEditor(filetype, desc.getId());
				}
			}
		}
	}

	/**
	 * Activates the Custom C Editor inside the SourceView
	 * and highlights the function selected on the CHView
	 * 
	 * @param region the function location 
	 */
	private void inputChanged(IRegion region) {
		// if the previous editor is active then dispose it
		if (editorpart != null) {
			editorpart.dispose();
			editorpart = null;
			top.dispose();
		}
		// set the current editor input
		resold = res;
		// create the editor
		createComposite();
		// reveal teh function selected in CHView
		if (editorpart instanceof ITextEditor) {
		    ITextEditor te= (ITextEditor) editorpart;
		    offset = region.getOffset();
		    length = region.getLength();
		    te.selectAndReveal(offset, length);
		}		
	}
	
	@Override
	public void dispose() {
		super.dispose();
		if(ref != null){
			((EditorReference)ref).dispose();
		}
		getSite().getWorkbenchWindow().getSelectionService().removeSelectionListener("org.eclipse.cdt.ui.callHierarchy", this);
	}

}
